#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <ctype.h>
#include <time.h>

#include <gtk/gtk.h>

#include "eggtrayicon.c"
#include "docklet.c"

#define	DEFAULT_ACTION_TIMEOUT	20
#define	DEFAULT_TIMEOUT		90000	/* 9 sec */
#define	DATA_PORT		33331
#define	CONF_PORT		33332

#ifndef __BSD_VISIBLE
char *strdup(char*);
#endif

int	timeout, action_timeout, last_action = 0;
char	*hostname, **global_env;
typedef struct t_list	LIST;
struct t_list {
	char		*item;
	LIST		*next;
};
LIST	**hosts = NULL;
LIST	**actions = NULL;
int	hcnt = 0, acnt = 0, is_alert = 0;

#define	INSERT(str,list)			\
{	LIST *t;				\
	t = (LIST*)malloc(sizeof(LIST));	\
	if (t == NULL) {			\
		fprintf(stderr, "OOM\n");	\
		exit(2);			\
	}					\
	t->item = strdup(str);			\
	t->next = list; list = t;		}
#define FOR_EACH(p,l)				\
	for (p = l; p != NULL; p = p->next)
struct t_service {
	char	*host;
	char	*serv;
	char	*stat;
	char	*info;
	time_t	updated;		
}	*serv = NULL;
int	serc = 0;
enum {
	COLUMN_HOST,
	COLUMN_SERVICE,
	COLUMN_STATUS,
	STATUS_COLOR,
	COLUMN_INFO,
	NUM_COLUMNS
};
GtkWidget	*window;
GdkColor	c0[3];

extern void set_icon(void);

void parse_config(void);
int parse_string(char *str, char **argv, char comment, int (*_isspace)(char));
void parse_data(char *data);
char *get_data(char *host, int port);

int iss(char c);

void minimize_to_tray(void)
{
    gtk_widget_hide(window);
    docklet_create();
}

void do_action(char *action)
{
	int		i, now;
	pid_t		auid;

	now = time(NULL);
	if ((now - last_action) < action_timeout)
		return;
	last_action = now;

	/* waitpid(-1, &s, 0);*/
	for (i = 0; i < acnt; i++) {
		if (!strcasecmp(actions[i]->item, action)) {
			auid = fork();
			if (auid == 0) {
				int	ac;
				char	*av[64];
				int	iss(char c) {
					if (c == ' ' || c == '\n')	
						return 1;
					return 0;
				}

				ac = parse_string(actions[i]->next->item, av, 0, iss);
				if (ac > 1) {
					if (!strcmp(av[0], "!")) {
						gtk_window_present(GTK_WINDOW(window));
						execve(av[1], av + 1, global_env);
					} else
    						execve(av[0], av, global_env);
				}
				exit(0);
			}
			if (auid > 0)
				return;
			fprintf(stderr, "Error on action %s\n", actions[i]->item);
		}			
	}
}
 
void create_model(GtkListStore	*store)
{
	char		*d;
	gint		i, j;
	GtkTreeIter	iter;
	GdkColor	*c1;
	LIST		*p, *l;
	
	if ((d = get_data(hostname, DATA_PORT)) != NULL) {

		parse_data(d);
		free(d);
		
		is_alert = 0;
		for (i = 0; i < serc; i++) {
			c1 = &c0[2];
			if (!strcmp(serv[i].stat, "CRITICAL"))
				c1 = &c0[0];
			else if (!strcmp(serv[i].stat, "WARNING"))
				c1 = &c0[1];
		
			gtk_list_store_append (store, &iter);
			gtk_list_store_set (store, &iter,
				  COLUMN_HOST,	 serv[i].host,
				  COLUMN_SERVICE,serv[i].serv,
				  COLUMN_STATUS, serv[i].stat,
				  STATUS_COLOR,	 c1,
				  COLUMN_INFO,   serv[i].info,
				  -1);

			/* lookup hostlist for matches */
			for (j = 0; j < hcnt; j++) {
				l = hosts[j]->next;
				FOR_EACH(p, l) {
					if (!strcasecmp(serv[i].host, p->item)) {
						is_alert = 1;
						do_action(hosts[j]->item);
						break;
					}
				}
			}
		}
		set_icon();
		
		if (serv != NULL) {
			for (i = 0; i < serc; i++) {
				free(serv[i].host);
				free(serv[i].serv);
				free(serv[i].stat);
				free(serv[i].info);
			}
			free(serv);
			serc = 0;
			serv = NULL;
		}
	}
}

static void add_columns (GtkTreeView *treeview)
{
	GtkCellRenderer		*renderer;
	GtkTreeViewColumn	*column;

	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Host", renderer, "text", COLUMN_HOST, NULL);
	gtk_tree_view_append_column (treeview, column);
	
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Service", renderer, "text", COLUMN_SERVICE, NULL);
	gtk_tree_view_append_column (treeview, column);
	
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Status", renderer, "text", COLUMN_STATUS, "background-gdk", STATUS_COLOR, NULL);
	gtk_tree_view_append_column (treeview, column);
	
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Info",  renderer, "text", COLUMN_INFO, NULL);
	gtk_tree_view_append_column (treeview, column);
}

gint update (gpointer data)
{
	GtkListStore		*model;
	
	gtk_tree_view_set_model(GTK_TREE_VIEW(data), NULL);
	model = gtk_list_store_new (NUM_COLUMNS,
			G_TYPE_STRING,
			G_TYPE_STRING,
			G_TYPE_STRING,
			GDK_TYPE_COLOR,
			G_TYPE_STRING);
	create_model(model);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model), COLUMN_STATUS, GTK_SORT_ASCENDING);
	gtk_tree_view_set_model(GTK_TREE_VIEW(data), GTK_TREE_MODEL(model));
	g_object_unref(model);

	return (TRUE);
}

int main (int argc, char **argv, char **envp)
{
	GtkWidget	*treeview, *vbox, *sw;
	gint quit(GtkWidget *widget, gpointer gdata) {
		gtk_main_quit();
		return (FALSE);
	}

	signal(SIGCHLD, SIG_IGN);
	global_env = envp;
	if (argc < 2) {
		fprintf(stderr, "Usage:\n%s <host>\n", argv[0]);
		return 2;
	}
	hostname = argv[1];
	parse_config();

	if (timeout < 500) {
		fprintf(stderr, "Timeout too small or undefined. Setting default.\n");
		timeout = DEFAULT_TIMEOUT;
	}
	if (action_timeout < 5) {
		fprintf(stderr, "Action timeout too small or undefined. Setting default.\n");
		action_timeout = DEFAULT_ACTION_TIMEOUT;
	}

	gtk_init(&argc, &argv);
	
	gdk_color_parse ("red",		&c0[0]);
	gdk_color_parse ("yellow",	&c0[1]);
	gdk_color_parse ("white",	&c0[2]);	

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window), "NagV");
    
	g_signal_connect (window, "delete_event", G_CALLBACK (minimize_to_tray), &window);
    
	gtk_container_set_border_width (GTK_CONTAINER (window), 8);
	gtk_window_set_default_size (GTK_WINDOW (window), 500, 250);
		
	vbox = gtk_vbox_new (FALSE, 8);
	gtk_container_add (GTK_CONTAINER (window), vbox);
	
	sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 0);		

	treeview = gtk_tree_view_new ();
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (treeview), TRUE);
	gtk_container_add (GTK_CONTAINER (sw), treeview);
	add_columns (GTK_TREE_VIEW (treeview));
	
	g_timeout_add(timeout, update, (gpointer)treeview);
	gtk_widget_show_all(window);
	
	gtk_main();
	return 0;
}

void parse_config(void)
{
	char	*c, *p, *config = get_data(hostname, CONF_PORT);
	void parse(char *s) {
		static int	line = 0;
		char		*av[64];
		int		ac, i;
		int iss(char c) {
			if (isspace(c))
				return 1;
			return 0;
		}
		
		if ((ac = parse_string(s, av, ';', iss)) < 1)
			return;
		if (!strcasecmp(av[0], "timeout")) {
			if (ac != 2) {
				fprintf(stderr, "%d: timeout requiers one argument: timeout <sec>\n", line);
				exit(2);
				return;
			}
			timeout = atoi(av[1]) * 1000;
			return;
		}
		if (!strcasecmp(av[0], "actiontimeout")) {
			if (ac != 2) {
				fprintf(stderr, "%d: actiontimeout requiers one argument: actiontimeout <sec>\n", line);
				exit(2);
				return;
			}
			action_timeout = atoi(av[1]);
			return;
		}
		if (!strcasecmp(av[0], "host-list")) {
			if (ac < 2) {
				fprintf(stderr, "%d: invalid host-list, usage:\n", line);
				fprintf(stderr, "%d:\thost-list <label> <host> ... <host>\n", line);
				return;
			}
			if (ac < 3) {
				fprintf(stderr, "%d: empty host-list, skip ...", line);
				return;
			}
			if ((hosts = (LIST**)realloc(hosts, (hcnt + 1) * sizeof(LIST*))) == NULL) {
				fprintf(stderr, "OOM\n");
				exit(2);
			}
			hosts[hcnt] = NULL;
			for (i = 2; i < ac; i++)
				INSERT(av[i], hosts[hcnt]);
			INSERT(av[1], hosts[hcnt]);
			hcnt++;
			return;
		}
		if (!strcasecmp(av[0], "action")) {
			char	*act = NULL;
			char *xstrcat(char *s1, char *s2) {
				int	l1, l2;
				char	*t;
				if (s1 == NULL && s2 == NULL)
					return NULL;
				if (s1 == NULL || (l1 = strlen(s1)) == 0)
					return strdup(s2);
				if (s2 == NULL || (l2 = strlen(s2)) == 0)
					return strdup(s1);
				t = (char*)malloc(l1 + l2 + 1);
				if (t == NULL) {
					fprintf(stderr, "xstrcat: OOM\n");
					exit(2);
				}
				memcpy(t, s1, l1);
				memcpy(t + l1, s2, l2);
				t[l1 + l2] = '\0';
				return t;
			}
			if (ac < 2) {
				fprintf(stderr, "%d: invalid action, usage:\n", line);
				fprintf(stderr, "%d:\taction <label> <action>\n", line);
				return;
			}
			if (ac < 3) {
				fprintf(stderr, "%d: empty action, skip ...", line);
				return;
			}
			if ((actions = (LIST**)realloc(actions, (acnt + 1) * sizeof(LIST*))) == NULL) {
				fprintf(stderr, "OOM\n");
				exit(2);
			}
			actions[acnt] = NULL;
			for (i = 2; i < ac; i++) {
				act = xstrcat(act, av[i]);
				act = xstrcat(act, " ");
			}
			INSERT(act, actions[acnt]);
			INSERT(av[1], actions[acnt]);
			free(act);
			acnt++;
			return;
		}
		fprintf(stderr, "%d: Uknown keyword '%s'\n", line, av[0]);
	}	

	if (config == NULL) {
		fprintf(stderr, "Unable to fetch config from %s:%d\n", hostname, CONF_PORT);
		exit(2);
	}

	for (c = config, p = c; *c != '\0'; c++)
		if (*c == '\n') {
			*c = '\0';
			parse(p);
			p = c + 1;
		}
	parse(p);
	free(config);
}

void parse_data(char *data)
{
	char	*c, *p;
	void parse(char *s) {
		char	*av[64];
		int	ac;
		int iss(char c) {
			if (c == ';' || c == '\n' || c == '\r')
				return 1;
			return 0;
		}
		
		if ((ac = parse_string(s, av, '#', iss)) != 32)
			return;
		if (strcmp(av[3], "OK")) {
			if ((serv = (struct t_service*)realloc(serv, (serc + 1) * sizeof(struct t_service))) == NULL) {
				fprintf(stderr, "OOM");
				return;
			}
			serv[serc].host = strdup(av[1]);
			serv[serc].serv = strdup(av[2]);
			serv[serc].stat = strdup(av[3]);
			serv[serc].info = strdup(av[31]);
			serc++;
		}
	}
	
	for (c = data, p = c; *c != '\0'; c++)
		if (*c == '\n') {
			*c = '\0';
			parse(p);
			p = c + 1;
		}
	parse(p);
}

char *get_data(char *host, int port)
{
	int			sockfd, numbytes, recbytes;
	char			buf[256], *r = NULL;
	struct hostent		*he;
	struct sockaddr_in	their_addr;	

	if ((he = gethostbyname(host)) == NULL) {
		perror("gethostbyname");
		return NULL;
	}
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		return NULL;
	}
	their_addr.sin_family = AF_INET;
	their_addr.sin_port = htons(port);
	their_addr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(their_addr.sin_zero), '\0', 8);
	if (connect(sockfd, (struct sockaddr *)&their_addr, sizeof(struct sockaddr)) == -1) {
		perror("connect");
		return NULL;
	}
	
	recbytes = 0;
	do {
		if ((numbytes = recv(sockfd, buf, 255, 0)) == -1) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			perror("recv");
			return NULL;
		}
		if ((r = (char*)realloc(r, recbytes + numbytes + 1)) == NULL) {
			fprintf(stderr, "OOM!\n");
			return NULL;
		}
		memcpy(r + recbytes, buf, numbytes);
		recbytes += numbytes;
	} while (numbytes > 0);
	r[recbytes] = '\0';
	close(sockfd);

	return r;
}

int parse_string(char *str, char **argv, char comment, int (*_isspace)(char))
{                                                                                                              
        int     argc = 0;                                                                                      
        char    *p;                                                                                            
                                                                                                               
        if (comment)                                                                                           
                if ((p = strchr(str, comment)) != NULL) {                                                      
                        *p = 0;                                                                                
                }                                                                                              
                                                                                                               
        while (argc < 63) {                                                                                    
                while (*str && _isspace(*str))
                        str++;                                                                                 
                if (*str == 0)                                                                                 
                        break;                                                                                 
                argv[argc++] = str;                                                                            
                while (*str && !_isspace(*str))
                        str++;                                                                                 
                if (*str)                                                                                      
                        *str++ = 0;                                                                            
        }                                                                                                      
        argv[argc] = NULL;                                                                                     
        return argc;                                                                                           
}
